/*
 * Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may not
 * use this file except in compliance with the License. A copy of the License
 * is located at
 *
 *    http://aws.amazon.com/apache2.0/
 *
 * or in the "license" file accompanying this file. This file is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
'use strict';

const CborEncoder = require('./CborEncoder');
const DaxClientError = require('./DaxClientError');
const DaxErrorCode = require('./DaxErrorCode');
const net = require('net');
const SessionVersion = require('./SessionVersion');
const SigV4Gen = require('./SigV4Gen');
const StreamBuffer = require('./ByteStreamBuffer');
const ControllablePromise = require('./ControllablePromise');

const MAGIC_STRING = 'J7yne5G';
const USER_AGENT_STRING = 'UserAgent';
const USER_AGENT = 'DaxJSClient-' + require('../package.json').version;
const WINDOW_SCALAR = 0.1;
const DAX_ADDR = 'https://dax.amazonaws.com';
const BUFFER_ZERO = Buffer.from([0]);
const DEFAULT_FLUSH_SIZE = 4096;

const MAX_PENDING_CONNECTION = 10; // same number as JAVA client
const ALLOC_TIMEOUT = 5000; // set timeout longer than JAVA client to avoid too many timeout caused by en/decoding inefficiency.

class TimeoutError extends Error {
  constructor(timeout) {
    super("Connection timeout after " + timeout + "ms");
  }
}

exports.USER_AGENT = USER_AGENT;

class ClientTube {
  constructor(socket, version, credProvider, region) {
    socket.setKeepAlive(true);
    socket.setNoDelay(true);
    this.socket = socket;
    this._authExp = 0;
    this._nextTube = null;
    this._authTTLMillis = 5 * 60 * 1000;
    this._poolWindow = (this._authTTLMillis / 2);
    this._tubeWindow = (this._authTTLMillis * WINDOW_SCALAR);
    this._region = region;
    this._credProvider = credProvider;
    this._sessionVersion = version;
    this.requestBuffer = new StreamBuffer();
    this.responseBuffer = new StreamBuffer();
    this.cbor = new CborEncoder();
    this._init(version.session);

    this._closed = false;
    this.socket.on('close', (had_error) => {
      // If the server closes the socket for some reason, mark the tube as closed
      this._closed = true;
    });
  }

  _init(session) {
    this.socket.write(this.cbor.encodeString(MAGIC_STRING));
    this.socket.write(BUFFER_ZERO);

    if(!session) {
      this.socket.write(this.cbor.encodeNull());
    } else {
      this.socket.write(this.cbor.encodeBinary(session));
    }

    this.socket.write(this.cbor.encodeMapHeader(1));

    this.socket.write(this.cbor.encodeString(USER_AGENT_STRING));
    this.socket.write(this.cbor.encodeString(USER_AGENT)); // FIXME

    this.socket.write(BUFFER_ZERO);
  }

  close() {
    this._closed = true;
    this.socket.end();
  }

  write(data) {
    this.requestBuffer.write(data);
    if(this.requestBuffer.length >= DEFAULT_FLUSH_SIZE) {
      this.socket.write(this.requestBuffer.read());
    }
  }

  flush(data) {
    if(this.requestBuffer.length > 0) {
      this.socket.write(this.requestBuffer.read());
    }
  }

  reauth() {
    let currTime = Date.now();
    if(this._authExp - currTime <= this._tubeWindow
        || currTime - this._lastPoolAuth >= this._poolWindow) {
      return this._credProvider.resolvePromise().then((creds) => {
        this._checkAndUpdateAccessKeyId(creds.accessKeyId);
        this._lastPoolAuth = currTime;
        this._authExp = currTime + this._authTTLMillis;

        this._authHandler(creds);

        // return the containing tube for further use
        return this;
      });
    } else {
      return Promise.resolve(this);
    }
  }

  invalidateAuth() {
    this._authExp = 0;
  }

  setTimeout(timeout, callback) {
    if(this.socket) {
      this.socket.setTimeout(timeout, callback);
    }
  }

  _authHandler(creds) {
    // Make sure the same credentials are used to sign and authorize.
    let sigHead = SigV4Gen.generateSigAndStringToSign(creds, DAX_ADDR, this._region, '');

    this.write(this.cbor.encodeInt(1));
    this.write(this.cbor.encodeInt(1489122155));
    this.write(this.cbor.encodeString(creds.accessKeyId));

    this.write(this.cbor.encodeString(sigHead.signature));

    this.write(this.cbor.encodeBinary(Buffer.from(sigHead.stringToSign)));

    this.write(sigHead.sessionToken === null ? this.cbor.encodeNull() : this.cbor.encodeString(sigHead.sessionToken));

    this.write(USER_AGENT === null ? this.cbor.encodeNull() : this.cbor.encodeString(USER_AGENT));
  }

  _checkAndUpdateAccessKeyId(other) {
    if(!other) {
      throw new DaxClientError('AWSCredentialsProvider provided null AWSAccessKeyId', DaxErrorCode.Authentication, false);
    }
    let equality = (other === this._accessKeyId);
    if(!equality) {
      this._accessKeyId = other;
    }
    return equality;
  }
}

class SocketTubePool {
  constructor(hostname, port, credProvider, region, idleTimeout, tube) {
    this._hostname = hostname;
    this._port = port;
    this._headTube = tube ? tube : null;

    this._region = region;
    this._credProvider = credProvider;
    this._sessionVersion = SessionVersion.create();

    this._pendingConnection = 0;
    this._pendingJob = [];

    this._idleTimeout = idleTimeout;
  }

  alloc() {
    let tube = this._headTube;
    while(tube !== null && tube._closed) {
      // find the first open tube, move it to the front
      this._headTube = tube = tube._nextTube;
    }

    if(tube != null) {
      // open tube is available, so use it
      this._headTube = tube._nextTube;
      tube._nextTube = null;
      if(tube.socket) {
        // remove the idle handler
        tube.socket.removeAllListeners("timeout"); 
        
        // ref socket before return to caller.
        tube.socket.ref();
      }

      return Promise.resolve(tube);
    } else {
      // no open, available tubes, so try to create one
      let wait = new ControllablePromise(ALLOC_TIMEOUT, new TimeoutError(ALLOC_TIMEOUT));
      this._pendingJob.push(wait);
      this._alloc();
      return wait;
    }
  }

  _alloc() {
    // separate this out since we can better unit test with mocking this func out.
    if(this._pendingConnection >= MAX_PENDING_CONNECTION) {
      return null;
    }

    this._pendingConnection++;
    let socket = net.createConnection(this._port, this._hostname, () => {
      let newTube = new ClientTube(socket, this._sessionVersion, this._credProvider, this._region);
      this.recycle(newTube);
      this._pendingConnection--;
    }).on('error', (err) => {
      this._pendingConnection--;
      console.error(err);
    });
  }

  recycle(tube) {
    if(!tube) {
      return;
    }

    if(tube._sessionVersion === this._sessionVersion) {
      if(tube.socket) {
        // first unref the socket recycled back to avoid hanging event loop.
        tube.socket.unref();

        // remove all socket listeners to avoid leaks
        tube.socket.removeAllListeners("timeout");
        tube.socket.removeAllListeners("data");
        tube.socket.removeAllListeners("error");
      }

      // first check whether we can assign tube to someone still waiting for it.
      while(this._pendingJob.length > 0) {
        let job = this._pendingJob.shift();
        if(job.isDone()) {
          continue;
        } else {
          if(tube.socket) {
            // ref socket before return to caller.
            tube.socket.ref();
          }
          job.resolve(tube);
          return;
        }
      }

      // no valid candidate, recycle to pool.
      // set the idle timeout to remove it if unused
      tube.setTimeout(this._idleTimeout, () => {
        this._removeIdleTube(tube);
      });
      tube._nextTube = this._headTube;
      this._headTube = tube;
    } else {
      tube.close();
    }
  }

  // when calling reset, it's most likely that all tubes are affected, so we
  // preemptively close every tube instead of waiting for each tube to get an
  // exception and closed.
  reset(tube) {
    if(!tube) {
      return;
    }
    tube.close();
    if(tube._sessionVersion !== this._sessionVersion) {
      return;
    }
    this._versionBump();
    tube = this._headTube;
    this._headTube = null;
    this._closeAll(tube);
  }

  close() {
    this._versionBump();
    let tube = this._headTube;
    this._headTube = null;
    this._closeAll(tube);
  }

  _closeAll(tube) {
    let reapCount = 0;
    let next;
    while(tube) {
      reapCount++;
      tube.close();
      next = tube._nextTube;
      tube._nextTube = null;
      tube = next;
    }
    return reapCount;
  }

  _versionBump() {
    this._sessionVersion = SessionVersion.create();
  }

  _removeIdleTube(tube) {
    if(!tube || !this._headTube) {
      return;
    }

    if(this._headTube === tube) {
      if(this._headTube._nextTube) {
        // if the head tube is idle and there's another tube available, remove the head tube
        this._headTube.close();
        this._headTube = this._headTube._nextTube;
      } else {
        // if there is no other tube, then leave it intact
        return;
      }
    } else {
      // find the idle tube in the list
      let prevTube = this._headTube, curTube = this._headTube._nextTube;
      while(curTube !== null) {
        if(curTube === tube) {
          // remove this tube from the list, let GC take care of it
          // prevTube cannot be null
          curTube.close();
          prevTube._nextTube = curTube._nextTube;
          return;
        } else {
          prevTube = curTube;
          curTube = curTube._nextTube;
        }
      }
    }

    // if we get here the tube was not found, but ignore it
  }
}

module.exports = {
  ClientTube: ClientTube,
  SocketTubePool: SocketTubePool,
  TimeoutError: TimeoutError,
};
